from deepdiff import DeepDiff

from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_tool_calling_node_inline_workflow.workflow import BasicToolCallingNodeInlineWorkflowWorkflow


def test_serialize_workflow():
    # GIVEN a Workflow that uses a generic node
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=BasicToolCallingNodeInlineWorkflowWorkflow)

    serialized_workflow: dict = workflow_display.serialize()
    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 1

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 2
    assert not DeepDiff(
        [
            {"id": "dbedc4ee-be3b-4135-8c26-3643c0b6a530", "key": "text", "type": "STRING"},
            {"id": "c5733df5-03bb-498e-a770-8ef9bff85df3", "key": "chat_history", "type": "CHAT_HISTORY"},
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    tool_calling_node = workflow_raw_data["nodes"][1]
    assert tool_calling_node == {
        "id": "21f29cac-da87-495f-bba1-093d423f4e46",
        "label": "GetCurrentWeatherNode",
        "type": "GENERIC",
        "display_data": {
            "position": {"x": 200.0, "y": -50.0},
            "comment": {"value": "\n    A tool calling node that calls the get_current_weather function.\n    "},
        },
        "base": {
            "name": "ToolCallingNode",
            "module": ["vellum", "workflows", "nodes", "displayable", "tool_calling_node", "node"],
        },
        "definition": {
            "name": "GetCurrentWeatherNode",
            "module": ["tests", "workflows", "basic_tool_calling_node_inline_workflow", "workflow"],
        },
        "trigger": {"id": "2414743b-b1dd-4552-8abf-9b7481df9762", "merge_behavior": "AWAIT_ATTRIBUTES"},
        "ports": [{"id": "3cd6d78c-9dad-42aa-ad38-31f67057c379", "name": "default", "type": "DEFAULT"}],
        "adornments": None,
        "attributes": [
            {
                "id": "44420e39-966f-4c59-bdf8-6365a61c5d2a",
                "name": "ml_model",
                "value": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "gpt-4o-mini"}},
            },
            {
                "id": "669cfb4b-8c25-460e-8952-b63d91302cbc",
                "name": "blocks",
                "value": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "JSON",
                        "value": [
                            {
                                "block_type": "CHAT_MESSAGE",
                                "state": None,
                                "cache_config": None,
                                "chat_role": "SYSTEM",
                                "chat_source": None,
                                "chat_message_unterminated": None,
                                "blocks": [
                                    {
                                        "block_type": "RICH_TEXT",
                                        "state": None,
                                        "cache_config": None,
                                        "blocks": [
                                            {
                                                "block_type": "PLAIN_TEXT",
                                                "state": None,
                                                "cache_config": None,
                                                "text": "You are a weather expert",
                                            }
                                        ],
                                    }
                                ],
                            },
                            {
                                "block_type": "CHAT_MESSAGE",
                                "state": None,
                                "cache_config": None,
                                "chat_role": "USER",
                                "chat_source": None,
                                "chat_message_unterminated": None,
                                "blocks": [
                                    {
                                        "block_type": "RICH_TEXT",
                                        "state": None,
                                        "cache_config": None,
                                        "blocks": [
                                            {
                                                "block_type": "VARIABLE",
                                                "state": None,
                                                "cache_config": None,
                                                "input_variable": "question",
                                            }
                                        ],
                                    }
                                ],
                            },
                        ],
                    },
                },
            },
            {
                "id": "78324739-ff89-47a5-902b-10da0cb95c6d",
                "name": "functions",
                "value": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "JSON",
                        "value": [
                            {
                                "type": "INLINE_WORKFLOW",
                                "name": "BasicInlineSubworkflowWorkflow",
                                "description": "\n    A workflow that gets the weather for a given city and date.\n    ",  # noqa: E501
                                "exec_config": {
                                    "workflow_raw_data": {
                                        "nodes": [
                                            {
                                                "id": "6358dcfe-b162-4e19-99ca-401d1ada9bdc",
                                                "type": "ENTRYPOINT",
                                                "inputs": [],
                                                "data": {
                                                    "label": "Entrypoint Node",
                                                    "source_handle_id": "c344fdee-282b-40c9-8c97-6dd08830948c",
                                                },
                                                "display_data": {"position": {"x": 0.0, "y": -50.0}},
                                                "base": None,
                                                "definition": None,
                                            },
                                            {
                                                "id": "1381c078-efa2-4255-89a1-7b4cb742c7fc",
                                                "label": "StartNode",
                                                "type": "GENERIC",
                                                "display_data": {"position": {"x": 200.0, "y": -50.0}},
                                                "base": {
                                                    "name": "BaseNode",
                                                    "module": ["vellum", "workflows", "nodes", "bases", "base"],
                                                },
                                                "definition": {
                                                    "name": "StartNode",
                                                    "module": [
                                                        "tests",
                                                        "workflows",
                                                        "basic_tool_calling_node_inline_workflow",
                                                        "workflow",
                                                    ],
                                                },
                                                "trigger": {
                                                    "id": "6492efcf-4437-4af1-9ad7-269795ccb27a",
                                                    "merge_behavior": "AWAIT_ATTRIBUTES",
                                                },
                                                "ports": [
                                                    {
                                                        "id": "1e739e86-a285-4438-9725-a152c15a63e3",
                                                        "name": "default",
                                                        "type": "DEFAULT",
                                                    }
                                                ],
                                                "adornments": None,
                                                "attributes": [
                                                    {
                                                        "id": "60ad78cd-fc78-4e08-926d-5a095b34d4f5",
                                                        "name": "city",
                                                        "value": {
                                                            "type": "WORKFLOW_INPUT",
                                                            "input_variable_id": "fa73da37-34c3-47a9-be58-69cc6cdbfca5",
                                                        },
                                                    },
                                                    {
                                                        "id": "c5f2d66c-5bb6-4d2a-8e4d-5356318cd3ba",
                                                        "name": "date",
                                                        "value": {
                                                            "type": "WORKFLOW_INPUT",
                                                            "input_variable_id": "aba1e6e0-dfa7-4c15-a4e6-aec6feebfaca",
                                                        },
                                                    },
                                                ],
                                                "outputs": [
                                                    {
                                                        "id": "3f4c753e-f057-47bb-9748-7968283cc8aa",
                                                        "name": "temperature",
                                                        "type": "NUMBER",
                                                        "value": None,
                                                    },
                                                    {
                                                        "id": "2a4a62b3-cd26-4d2c-b3f1-eaa5f9dd22dd",
                                                        "name": "reasoning",
                                                        "type": "STRING",
                                                        "value": None,
                                                    },
                                                ],
                                            },
                                            {
                                                "id": "0779b232-82ab-4dbe-a340-6a85e6ab3368",
                                                "type": "TERMINAL",
                                                "data": {
                                                    "label": "Final Output",
                                                    "name": "temperature",
                                                    "target_handle_id": "9e077063-c394-4c7b-b0c6-e6686df67984",
                                                    "output_id": "99afb757-2782-465d-ab55-80ccf50552b9",
                                                    "output_type": "NUMBER",
                                                    "node_input_id": "7761c5e1-cc2e-43ab-bfd2-f66c3d47b3b9",
                                                },
                                                "inputs": [
                                                    {
                                                        "id": "7761c5e1-cc2e-43ab-bfd2-f66c3d47b3b9",
                                                        "key": "node_input",
                                                        "value": {
                                                            "rules": [
                                                                {
                                                                    "type": "NODE_OUTPUT",
                                                                    "data": {
                                                                        "node_id": "1381c078-efa2-4255-89a1-7b4cb742c7fc",  # noqa: E501
                                                                        "output_id": "3f4c753e-f057-47bb-9748-7968283cc8aa",  # noqa: E501
                                                                    },
                                                                }
                                                            ],
                                                            "combinator": "OR",
                                                        },
                                                    }
                                                ],
                                                "display_data": {"position": {"x": 400.0, "y": -175.0}},
                                                "base": {
                                                    "name": "FinalOutputNode",
                                                    "module": [
                                                        "vellum",
                                                        "workflows",
                                                        "nodes",
                                                        "displayable",
                                                        "final_output_node",
                                                        "node",
                                                    ],
                                                },
                                                "definition": None,
                                            },
                                            {
                                                "id": "31b74695-3f1c-47cf-8be8-a4d86cc589e8",
                                                "type": "TERMINAL",
                                                "data": {
                                                    "label": "Final Output",
                                                    "name": "reasoning",
                                                    "target_handle_id": "8b525943-6c27-414b-a329-e29c0b217f72",
                                                    "output_id": "7444a019-081a-4e10-a528-3249299159f7",
                                                    "output_type": "STRING",
                                                    "node_input_id": "c1833b54-95b6-4365-8e57-51b09c8e2606",
                                                },
                                                "inputs": [
                                                    {
                                                        "id": "c1833b54-95b6-4365-8e57-51b09c8e2606",
                                                        "key": "node_input",
                                                        "value": {
                                                            "rules": [
                                                                {
                                                                    "type": "NODE_OUTPUT",
                                                                    "data": {
                                                                        "node_id": "1381c078-efa2-4255-89a1-7b4cb742c7fc",  # noqa: E501
                                                                        "output_id": "2a4a62b3-cd26-4d2c-b3f1-eaa5f9dd22dd",  # noqa: E501
                                                                    },
                                                                }
                                                            ],
                                                            "combinator": "OR",
                                                        },
                                                    }
                                                ],
                                                "display_data": {"position": {"x": 400.0, "y": 75.0}},
                                                "base": {
                                                    "name": "FinalOutputNode",
                                                    "module": [
                                                        "vellum",
                                                        "workflows",
                                                        "nodes",
                                                        "displayable",
                                                        "final_output_node",
                                                        "node",
                                                    ],
                                                },
                                                "definition": None,
                                            },
                                        ],
                                        "edges": [
                                            {
                                                "id": "a37781d1-f7a5-4386-a67d-0c3d5c929602",
                                                "source_node_id": "6358dcfe-b162-4e19-99ca-401d1ada9bdc",
                                                "source_handle_id": "c344fdee-282b-40c9-8c97-6dd08830948c",
                                                "target_node_id": "1381c078-efa2-4255-89a1-7b4cb742c7fc",
                                                "target_handle_id": "6492efcf-4437-4af1-9ad7-269795ccb27a",
                                                "type": "DEFAULT",
                                            },
                                            {
                                                "id": "3c5d8990-48f5-42e1-893e-bc8308d2110a",
                                                "source_node_id": "1381c078-efa2-4255-89a1-7b4cb742c7fc",
                                                "source_handle_id": "1e739e86-a285-4438-9725-a152c15a63e3",
                                                "target_node_id": "0779b232-82ab-4dbe-a340-6a85e6ab3368",
                                                "target_handle_id": "9e077063-c394-4c7b-b0c6-e6686df67984",
                                                "type": "DEFAULT",
                                            },
                                            {
                                                "id": "de0b8090-a26e-4e09-9173-9f7400a5be4c",
                                                "source_node_id": "1381c078-efa2-4255-89a1-7b4cb742c7fc",
                                                "source_handle_id": "1e739e86-a285-4438-9725-a152c15a63e3",
                                                "target_node_id": "31b74695-3f1c-47cf-8be8-a4d86cc589e8",
                                                "target_handle_id": "8b525943-6c27-414b-a329-e29c0b217f72",
                                                "type": "DEFAULT",
                                            },
                                        ],
                                        "display_data": {"viewport": {"x": 0.0, "y": 0.0, "zoom": 1.0}},
                                        "definition": {
                                            "name": "BasicInlineSubworkflowWorkflow",
                                            "module": [
                                                "tests",
                                                "workflows",
                                                "basic_tool_calling_node_inline_workflow",
                                                "workflow",
                                            ],
                                        },
                                        "output_values": [
                                            {
                                                "output_variable_id": "99afb757-2782-465d-ab55-80ccf50552b9",
                                                "value": {
                                                    "type": "NODE_OUTPUT",
                                                    "node_id": "1381c078-efa2-4255-89a1-7b4cb742c7fc",
                                                    "node_output_id": "3f4c753e-f057-47bb-9748-7968283cc8aa",
                                                },
                                            },
                                            {
                                                "output_variable_id": "7444a019-081a-4e10-a528-3249299159f7",
                                                "value": {
                                                    "type": "NODE_OUTPUT",
                                                    "node_id": "1381c078-efa2-4255-89a1-7b4cb742c7fc",
                                                    "node_output_id": "2a4a62b3-cd26-4d2c-b3f1-eaa5f9dd22dd",
                                                },
                                            },
                                        ],
                                    },
                                    "input_variables": [
                                        {
                                            "id": "fa73da37-34c3-47a9-be58-69cc6cdbfca5",
                                            "key": "city",
                                            "type": "STRING",
                                            "default": None,
                                            "required": True,
                                            "extensions": {"color": None},
                                        },
                                        {
                                            "id": "aba1e6e0-dfa7-4c15-a4e6-aec6feebfaca",
                                            "key": "date",
                                            "type": "STRING",
                                            "default": None,
                                            "required": True,
                                            "extensions": {"color": None},
                                        },
                                    ],
                                    "state_variables": [],
                                    "output_variables": [
                                        {
                                            "id": "99afb757-2782-465d-ab55-80ccf50552b9",
                                            "key": "temperature",
                                            "type": "NUMBER",
                                        },
                                        {
                                            "id": "7444a019-081a-4e10-a528-3249299159f7",
                                            "key": "reasoning",
                                            "type": "STRING",
                                        },
                                    ],
                                },
                            }
                        ],
                    },
                },
            },
            {
                "id": "0f6dc102-3460-4963-91fa-7ba85d65ef7a",
                "name": "prompt_inputs",
                "value": {
                    "type": "DICTIONARY_REFERENCE",
                    "entries": [
                        {
                            "id": "d5c4d578-6ef1-4786-88f6-1ab0892d0798",
                            "key": "question",
                            "value": {
                                "type": "WORKFLOW_INPUT",
                                "input_variable_id": "045942b7-e5b9-482c-b4d4-943309a20e05",
                            },
                        }
                    ],
                },
            },
            {
                "id": "229cd1ca-dc2f-4586-b933-c4d4966f7bd1",
                "name": "parameters",
                "value": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "JSON",
                        "value": {
                            "stop": [],
                            "temperature": 0.0,
                            "max_tokens": 4096.0,
                            "top_p": 1.0,
                            "top_k": 0.0,
                            "frequency_penalty": 0.0,
                            "presence_penalty": 0.0,
                            "logit_bias": None,
                            "custom_parameters": None,
                        },
                    },
                },
            },
            {
                "id": "1668419e-a193-43a5-8a97-3394e89bf278",
                "name": "max_prompt_iterations",
                "value": {"type": "CONSTANT_VALUE", "value": {"type": "NUMBER", "value": 5.0}},
            },
        ],
        "outputs": [
            {"id": "e62bc785-a914-4066-b79e-8c89a5d0ec6c", "name": "text", "type": "STRING", "value": None},
            {
                "id": "4674f1d9-e3af-411f-8a55-40a3a3ab5394",
                "name": "chat_history",
                "type": "CHAT_HISTORY",
                "value": None,
            },
        ],
    }



# This file was auto-generated by Fern from our API Definition.

import typing
from ...core.client_wrapper import SyncClientWrapper
from .raw_client import RawWorkflowDeploymentsClient
from .types.workflow_deployments_list_request_status import WorkflowDeploymentsListRequestStatus
from ...core.request_options import RequestOptions
from ...types.paginated_slim_workflow_deployment_list import PaginatedSlimWorkflowDeploymentList
from ...types.workflow_deployment_read import WorkflowDeploymentRead
from ...types.workflow_deployment_event_executions_response import WorkflowDeploymentEventExecutionsResponse
from ...types.workflow_event_execution_read import WorkflowEventExecutionRead
from ...types.workflow_deployment_history_item import WorkflowDeploymentHistoryItem
from .types.list_workflow_release_tags_request_source import ListWorkflowReleaseTagsRequestSource
from ...types.paginated_workflow_release_tag_read_list import PaginatedWorkflowReleaseTagReadList
from ...types.workflow_release_tag_read import WorkflowReleaseTagRead
from ...types.workflow_deployment_release import WorkflowDeploymentRelease
from ...core.client_wrapper import AsyncClientWrapper
from .raw_client import AsyncRawWorkflowDeploymentsClient

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class WorkflowDeploymentsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawWorkflowDeploymentsClient(client_wrapper=client_wrapper)

    @property
    def _client_wrapper(self) -> SyncClientWrapper:
        return self._raw_client._client_wrapper

    @property
    def with_raw_response(self) -> RawWorkflowDeploymentsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawWorkflowDeploymentsClient
        """
        return self._raw_client

    def list(
        self,
        *,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        status: typing.Optional[WorkflowDeploymentsListRequestStatus] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PaginatedSlimWorkflowDeploymentList:
        """
        Used to list all Workflow Deployments.

        Parameters
        ----------
        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        status : typing.Optional[WorkflowDeploymentsListRequestStatus]
            status

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PaginatedSlimWorkflowDeploymentList


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.workflow_deployments.list()
        """
        response = self._raw_client.list(
            limit=limit, offset=offset, ordering=ordering, status=status, request_options=request_options
        )
        return response.data

    def retrieve(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> WorkflowDeploymentRead:
        """
        Used to retrieve a workflow deployment given its ID or name.

        Parameters
        ----------
        id : str
            Either the Workflow Deployment's ID or its unique name

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowDeploymentRead


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.workflow_deployments.retrieve(
            id="id",
        )
        """
        response = self._raw_client.retrieve(id, request_options=request_options)
        return response.data

    def list_workflow_deployment_event_executions(
        self,
        id: str,
        *,
        filters: typing.Optional[str] = None,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> WorkflowDeploymentEventExecutionsResponse:
        """
        Parameters
        ----------
        id : str

        filters : typing.Optional[str]

        limit : typing.Optional[int]
            Number of executions to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the executions.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowDeploymentEventExecutionsResponse


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.workflow_deployments.list_workflow_deployment_event_executions(
            id="id",
        )
        """
        response = self._raw_client.list_workflow_deployment_event_executions(
            id, filters=filters, limit=limit, offset=offset, request_options=request_options
        )
        return response.data

    def workflow_deployment_event_execution(
        self, execution_id: str, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> WorkflowEventExecutionRead:
        """
        Parameters
        ----------
        execution_id : str

        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowEventExecutionRead


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.workflow_deployments.workflow_deployment_event_execution(
            execution_id="execution_id",
            id="id",
        )
        """
        response = self._raw_client.workflow_deployment_event_execution(
            execution_id, id, request_options=request_options
        )
        return response.data

    def workflow_deployment_history_item_retrieve(
        self, history_id_or_release_tag: str, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> WorkflowDeploymentHistoryItem:
        """
        DEPRECATED: This endpoint is deprecated and will be removed in a future release. Please use the
        `retrieve_workflow_deployment_release` endpoint instead.

        Parameters
        ----------
        history_id_or_release_tag : str
            Either the UUID of Workflow Deployment History Item you'd like to retrieve, or the name of a Release Tag that's pointing to the Workflow Deployment History Item you'd like to retrieve.

        id : str
            Either the Workflow Deployment's ID or its unique name

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowDeploymentHistoryItem


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.workflow_deployments.workflow_deployment_history_item_retrieve(
            history_id_or_release_tag="history_id_or_release_tag",
            id="id",
        )
        """
        response = self._raw_client.workflow_deployment_history_item_retrieve(
            history_id_or_release_tag, id, request_options=request_options
        )
        return response.data

    def list_workflow_release_tags(
        self,
        id: str,
        *,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        source: typing.Optional[ListWorkflowReleaseTagsRequestSource] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PaginatedWorkflowReleaseTagReadList:
        """
        List Release Tags associated with the specified Workflow Deployment

        Parameters
        ----------
        id : str
            Either the Workflow Deployment's ID or its unique name

        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        source : typing.Optional[ListWorkflowReleaseTagsRequestSource]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PaginatedWorkflowReleaseTagReadList


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.workflow_deployments.list_workflow_release_tags(
            id="id",
        )
        """
        response = self._raw_client.list_workflow_release_tags(
            id, limit=limit, offset=offset, ordering=ordering, source=source, request_options=request_options
        )
        return response.data

    def retrieve_workflow_release_tag(
        self, id: str, name: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> WorkflowReleaseTagRead:
        """
        Retrieve a Workflow Release Tag by tag name, associated with a specified Workflow Deployment.

        Parameters
        ----------
        id : str
            Either the Workflow Deployment's ID or its unique name

        name : str
            The name of the Release Tag associated with this Workflow Deployment that you'd like to retrieve.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowReleaseTagRead


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.workflow_deployments.retrieve_workflow_release_tag(
            id="id",
            name="name",
        )
        """
        response = self._raw_client.retrieve_workflow_release_tag(id, name, request_options=request_options)
        return response.data

    def update_workflow_release_tag(
        self,
        id: str,
        name: str,
        *,
        history_item_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> WorkflowReleaseTagRead:
        """
        Updates an existing Release Tag associated with the specified Workflow Deployment.

        Parameters
        ----------
        id : str
            Either the Workflow Deployment's ID or its unique name

        name : str
            The name of the Release Tag associated with this Workflow Deployment that you'd like to update.

        history_item_id : typing.Optional[str]
            The ID of the Workflow Deployment History Item to tag

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowReleaseTagRead


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.workflow_deployments.update_workflow_release_tag(
            id="id",
            name="name",
        )
        """
        response = self._raw_client.update_workflow_release_tag(
            id, name, history_item_id=history_item_id, request_options=request_options
        )
        return response.data

    def retrieve_workflow_deployment_release(
        self, id: str, release_id_or_release_tag: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> WorkflowDeploymentRelease:
        """
        Retrieve a specific Workflow Deployment Release by either its UUID or the name of a Release Tag that points to it.

        Parameters
        ----------
        id : str
            Either the Workflow Deployment's ID or its unique name

        release_id_or_release_tag : str
            Either the UUID of Workflow Deployment Release you'd like to retrieve, or the name of a Release Tag that's pointing to the Workflow Deployment Release you'd like to retrieve.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowDeploymentRelease


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.workflow_deployments.retrieve_workflow_deployment_release(
            id="id",
            release_id_or_release_tag="release_id_or_release_tag",
        )
        """
        response = self._raw_client.retrieve_workflow_deployment_release(
            id, release_id_or_release_tag, request_options=request_options
        )
        return response.data


class AsyncWorkflowDeploymentsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawWorkflowDeploymentsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawWorkflowDeploymentsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawWorkflowDeploymentsClient
        """
        return self._raw_client

    async def list(
        self,
        *,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        status: typing.Optional[WorkflowDeploymentsListRequestStatus] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PaginatedSlimWorkflowDeploymentList:
        """
        Used to list all Workflow Deployments.

        Parameters
        ----------
        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        status : typing.Optional[WorkflowDeploymentsListRequestStatus]
            status

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PaginatedSlimWorkflowDeploymentList


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.workflow_deployments.list()


        asyncio.run(main())
        """
        response = await self._raw_client.list(
            limit=limit, offset=offset, ordering=ordering, status=status, request_options=request_options
        )
        return response.data

    async def retrieve(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> WorkflowDeploymentRead:
        """
        Used to retrieve a workflow deployment given its ID or name.

        Parameters
        ----------
        id : str
            Either the Workflow Deployment's ID or its unique name

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowDeploymentRead


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.workflow_deployments.retrieve(
                id="id",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.retrieve(id, request_options=request_options)
        return response.data

    async def list_workflow_deployment_event_executions(
        self,
        id: str,
        *,
        filters: typing.Optional[str] = None,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> WorkflowDeploymentEventExecutionsResponse:
        """
        Parameters
        ----------
        id : str

        filters : typing.Optional[str]

        limit : typing.Optional[int]
            Number of executions to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the executions.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowDeploymentEventExecutionsResponse


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.workflow_deployments.list_workflow_deployment_event_executions(
                id="id",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.list_workflow_deployment_event_executions(
            id, filters=filters, limit=limit, offset=offset, request_options=request_options
        )
        return response.data

    async def workflow_deployment_event_execution(
        self, execution_id: str, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> WorkflowEventExecutionRead:
        """
        Parameters
        ----------
        execution_id : str

        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowEventExecutionRead


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.workflow_deployments.workflow_deployment_event_execution(
                execution_id="execution_id",
                id="id",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.workflow_deployment_event_execution(
            execution_id, id, request_options=request_options
        )
        return response.data

    async def workflow_deployment_history_item_retrieve(
        self, history_id_or_release_tag: str, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> WorkflowDeploymentHistoryItem:
        """
        DEPRECATED: This endpoint is deprecated and will be removed in a future release. Please use the
        `retrieve_workflow_deployment_release` endpoint instead.

        Parameters
        ----------
        history_id_or_release_tag : str
            Either the UUID of Workflow Deployment History Item you'd like to retrieve, or the name of a Release Tag that's pointing to the Workflow Deployment History Item you'd like to retrieve.

        id : str
            Either the Workflow Deployment's ID or its unique name

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowDeploymentHistoryItem


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.workflow_deployments.workflow_deployment_history_item_retrieve(
                history_id_or_release_tag="history_id_or_release_tag",
                id="id",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.workflow_deployment_history_item_retrieve(
            history_id_or_release_tag, id, request_options=request_options
        )
        return response.data

    async def list_workflow_release_tags(
        self,
        id: str,
        *,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        source: typing.Optional[ListWorkflowReleaseTagsRequestSource] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PaginatedWorkflowReleaseTagReadList:
        """
        List Release Tags associated with the specified Workflow Deployment

        Parameters
        ----------
        id : str
            Either the Workflow Deployment's ID or its unique name

        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        source : typing.Optional[ListWorkflowReleaseTagsRequestSource]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PaginatedWorkflowReleaseTagReadList


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.workflow_deployments.list_workflow_release_tags(
                id="id",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.list_workflow_release_tags(
            id, limit=limit, offset=offset, ordering=ordering, source=source, request_options=request_options
        )
        return response.data

    async def retrieve_workflow_release_tag(
        self, id: str, name: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> WorkflowReleaseTagRead:
        """
        Retrieve a Workflow Release Tag by tag name, associated with a specified Workflow Deployment.

        Parameters
        ----------
        id : str
            Either the Workflow Deployment's ID or its unique name

        name : str
            The name of the Release Tag associated with this Workflow Deployment that you'd like to retrieve.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowReleaseTagRead


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.workflow_deployments.retrieve_workflow_release_tag(
                id="id",
                name="name",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.retrieve_workflow_release_tag(id, name, request_options=request_options)
        return response.data

    async def update_workflow_release_tag(
        self,
        id: str,
        name: str,
        *,
        history_item_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> WorkflowReleaseTagRead:
        """
        Updates an existing Release Tag associated with the specified Workflow Deployment.

        Parameters
        ----------
        id : str
            Either the Workflow Deployment's ID or its unique name

        name : str
            The name of the Release Tag associated with this Workflow Deployment that you'd like to update.

        history_item_id : typing.Optional[str]
            The ID of the Workflow Deployment History Item to tag

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowReleaseTagRead


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.workflow_deployments.update_workflow_release_tag(
                id="id",
                name="name",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.update_workflow_release_tag(
            id, name, history_item_id=history_item_id, request_options=request_options
        )
        return response.data

    async def retrieve_workflow_deployment_release(
        self, id: str, release_id_or_release_tag: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> WorkflowDeploymentRelease:
        """
        Retrieve a specific Workflow Deployment Release by either its UUID or the name of a Release Tag that points to it.

        Parameters
        ----------
        id : str
            Either the Workflow Deployment's ID or its unique name

        release_id_or_release_tag : str
            Either the UUID of Workflow Deployment Release you'd like to retrieve, or the name of a Release Tag that's pointing to the Workflow Deployment Release you'd like to retrieve.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowDeploymentRelease


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.workflow_deployments.retrieve_workflow_deployment_release(
                id="id",
                release_id_or_release_tag="release_id_or_release_tag",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.retrieve_workflow_deployment_release(
            id, release_id_or_release_tag, request_options=request_options
        )
        return response.data



from typing import Any, Dict, cast

from vellum.client.types.vellum_error import VellumError
from vellum.workflows import BaseWorkflow
from vellum.workflows.nodes.core.error_node.node import ErrorNode
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


def test_error_node_display__serialize_with_vellum_error() -> None:
    # GIVEN an Error Node with a VellumError
    class MyNode(ErrorNode):
        error = VellumError(
            message="A bad thing happened",
            code="USER_DEFINED_ERROR",
        )

    # AND a workflow referencing the two node
    class MyWorkflow(BaseWorkflow):
        graph = MyNode

    # WHEN we serialize the workflow
    workflow_display = get_workflow_display(workflow_class=MyWorkflow)
    serialized_workflow = cast(Dict[str, Any], workflow_display.serialize())

    # THEN the correct inputs should be serialized on the node
    serialized_node = next(
        node for node in serialized_workflow["workflow_raw_data"]["nodes"] if node["id"] == str(MyNode.__id__)
    )
    assert serialized_node["inputs"][0]["value"] == {
        "combinator": "OR",
        "rules": [
            {
                "data": {
                    "type": "ERROR",
                    "value": {
                        "message": "A bad thing happened",
                        "code": "USER_DEFINED_ERROR",
                    },
                },
                "type": "CONSTANT_VALUE",
            }
        ],
    }

    # AND we serialize the DEPRECATED fields
    assert "error_output_id" not in serialized_node["data"]
    assert "name" not in serialized_node["data"]



from typing import TYPE_CHECKING, Any, Iterator, List, Optional, Type

from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema

from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.descriptors.exceptions import InvalidExpressionException
from vellum.workflows.edges.edge import Edge
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException, WorkflowInitializationException
from vellum.workflows.graph import Graph, GraphTarget
from vellum.workflows.state.base import BaseState
from vellum.workflows.types.core import ConditionType

if TYPE_CHECKING:
    from vellum.workflows.nodes.bases import BaseNode


class Port:
    node_class: Type["BaseNode"]

    _edges: List[Edge]
    _condition: Optional[BaseDescriptor]
    _condition_type: Optional[ConditionType]

    def __init__(
        self,
        default: bool = False,
        fork_state: bool = False,
        condition: Optional[Any] = None,
        condition_type: Optional[ConditionType] = None,
    ):
        self.default = default
        self.node_class = None  # type: ignore[assignment]
        self._fork_state = fork_state
        self._edges = []
        self._condition: Optional[BaseDescriptor] = condition
        self._condition_type: Optional[ConditionType] = condition_type

    def __set_name__(self, owner: Type, name: str) -> None:
        self.name = name

    def __repr__(self) -> str:
        return f"{self.node_class}.Ports.{self.name}"

    def copy(self) -> "Port":
        return Port(
            default=self.default,
            fork_state=self._fork_state,
            condition=self._condition,
            condition_type=self._condition_type,
        )

    @property
    def fork_state(self) -> bool:
        return self._fork_state

    @property
    def edges(self) -> Iterator[Edge]:
        return iter(self._edges)

    def __rshift__(self, other: GraphTarget) -> Graph:
        if isinstance(other, set) or isinstance(other, Graph):
            return Graph.from_port(self) >> other

        if isinstance(other, Port):
            return Graph.from_port(self) >> Graph.from_port(other)

        edge = Edge(from_port=self, to_node=other)
        if edge not in self._edges:
            self._edges.append(edge)

        return Graph.from_edge(edge)

    @staticmethod
    def on_if(condition: Optional[BaseDescriptor] = None, fork_state: bool = False):
        return Port(condition=condition, condition_type=ConditionType.IF, fork_state=fork_state)

    @staticmethod
    def on_elif(condition: Optional[BaseDescriptor] = None, fork_state: bool = False) -> "Port":
        return Port(condition=condition, condition_type=ConditionType.ELIF, fork_state=fork_state)

    @staticmethod
    def on_else(fork_state: bool = False) -> "Port":
        return Port(condition_type=ConditionType.ELSE, fork_state=fork_state)

    def resolve_condition(self, state: BaseState) -> bool:
        try:
            if self._condition is None:
                return False

            value = self._condition.resolve(state)
            return bool(value)
        except InvalidExpressionException as e:
            raise NodeException(
                message=f"Failed to resolve condition for port `{self.name}`: {e}",
                code=WorkflowErrorCode.INVALID_INPUTS,
            ) from e

    def serialize(self) -> dict:
        return {
            "name": self.name,
        }

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.is_instance_schema(cls)

    def validate(self):
        if (
            not self.default
            and self._condition_type in (ConditionType.IF, ConditionType.ELIF)
            and self._condition is None
        ):
            raise WorkflowInitializationException(
                f"Class {self.node_class.__name__}'s {self.name} should have a defined condition and cannot be empty."
            )



import pytest
import logging
from uuid import UUID, uuid4

from vellum.workflows.edges.edge import Edge
from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.nodes.core.inline_subworkflow_node.node import InlineSubworkflowNode
from vellum.workflows.outputs.base import BaseOutputs
from vellum.workflows.state.base import BaseState
from vellum.workflows.types.stack import Stack
from vellum.workflows.workflows.base import BaseWorkflow


class GlobalTestWorkflow(BaseWorkflow):
    """
    Used as part of `test_base_workflow__deserialize_state_with_invalid_workflow_definition` below.
    """

    pass


def test_base_workflow__inherit_base_outputs():
    class MyNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

        def run(self):
            return self.Outputs(foo="bar")

    class MyWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = MyNode

        class Outputs:
            output = MyNode.Outputs.foo

    # TEST that the Outputs class is a subclass of BaseOutputs
    assert issubclass(MyWorkflow.Outputs, BaseOutputs)

    # TEST that the Outputs class does not inherit from object
    assert object not in MyWorkflow.Outputs.__bases__

    workflow = MyWorkflow()
    terminal_event = workflow.run()

    # TEST that the Outputs class has the correct attributes
    assert hasattr(MyWorkflow.Outputs, "output")

    # TEST that the outputs should be correct
    assert terminal_event.name == "workflow.execution.fulfilled", terminal_event
    assert terminal_event.outputs == {"output": "bar"}


def test_subworkflow__inherit_base_outputs():
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

        def run(self):
            return self.Outputs(foo="bar")

    class SubWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = StartNode

        class Outputs:
            output = StartNode.Outputs.foo

    class SubworkflowNode(InlineSubworkflowNode):
        subworkflow = SubWorkflow

    class MainWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = SubworkflowNode

        class Outputs:
            output = SubworkflowNode.Outputs.output

    # TEST that the Outputs classes are subclasses of BaseOutputs
    assert issubclass(MainWorkflow.Outputs, BaseOutputs)
    assert issubclass(SubWorkflow.Outputs, BaseOutputs)

    # TEST that the Outputs classes do not inherit from object
    assert object not in MainWorkflow.Outputs.__bases__

    # TEST execution
    workflow = MainWorkflow()
    terminal_event = workflow.run()

    # TEST that the Outputs class has the correct attributes
    assert hasattr(MainWorkflow.Outputs, "output")

    # TEST that the outputs are correct
    assert terminal_event.name == "workflow.execution.fulfilled", terminal_event
    assert terminal_event.outputs == {"output": "bar"}


def test_workflow__single_node():
    class NodeA(BaseNode):
        pass

    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = NodeA

    nodes = set(TestWorkflow.get_nodes())
    assert nodes == {NodeA}


def test_workflow__multiple_nodes():
    class NodeA(BaseNode):
        pass

    class NodeB(BaseNode):
        pass

    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = {NodeA, NodeB}

    nodes = set(TestWorkflow.get_nodes())
    assert nodes == {NodeA, NodeB}


def test_workflow__single_graph():
    class NodeA(BaseNode):
        pass

    class NodeB(BaseNode):
        pass

    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = NodeA >> NodeB

    nodes = set(TestWorkflow.get_nodes())
    assert nodes == {NodeA, NodeB}


def test_workflow__multiple_graphs():
    class NodeA(BaseNode):
        pass

    class NodeB(BaseNode):
        pass

    class NodeC(BaseNode):
        pass

    class NodeD(BaseNode):
        pass

    class NodeE(BaseNode):
        pass

    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = NodeA >> {NodeB >> NodeC, NodeD} >> NodeE

    nodes = set(TestWorkflow.get_nodes())
    assert nodes == {NodeA, NodeB, NodeC, NodeD, NodeE}


def test_workflow__get_edges():
    class NodeA(BaseNode):
        pass

    class NodeB(BaseNode):
        pass

    class NodeC(BaseNode):
        pass

    class NodeD(BaseNode):
        pass

    class NodeE(BaseNode):
        pass

    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = {NodeA >> {NodeB >> NodeC, NodeD} >> NodeE}

    edges = set(TestWorkflow.get_edges())
    assert edges == {
        Edge(from_port=NodeA.Ports.default, to_node=NodeB),
        Edge(from_port=NodeB.Ports.default, to_node=NodeC),
        Edge(from_port=NodeA.Ports.default, to_node=NodeD),
        Edge(from_port=NodeC.Ports.default, to_node=NodeE),
        Edge(from_port=NodeD.Ports.default, to_node=NodeE),
    }


def test_workflow__get_unused_nodes():
    class NodeA(BaseNode):
        pass

    class NodeB(BaseNode):
        pass

    class NodeC(BaseNode):
        pass

    class NodeD(BaseNode):
        pass

    class NodeE(BaseNode):
        pass

    class NodeF(BaseNode):
        pass

    class NodeG(BaseNode):
        pass

    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = NodeA
        unused_graphs = {NodeB >> NodeC, NodeD >> {NodeE >> NodeF, NodeG}}

    unused_graphs = set(TestWorkflow.get_unused_nodes())
    assert unused_graphs == {NodeB, NodeC, NodeD, NodeE, NodeF, NodeG}


def test_workflow__get_unused_edges():
    """
    Test that get_unused_edges correctly identifies edges that are defined but not used in the workflow graph.
    """

    class NodeA(BaseNode):
        pass

    class NodeB(BaseNode):
        pass

    class NodeC(BaseNode):
        pass

    class NodeD(BaseNode):
        pass

    class NodeE(BaseNode):
        pass

    class NodeF(BaseNode):
        pass

    class NodeG(BaseNode):
        pass

    class NodeH(BaseNode):
        pass

    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = NodeA >> NodeB
        unused_graphs = {NodeC >> NodeD, NodeE >> {NodeF, NodeG} >> NodeH}

    # Collect unused edges
    unused_edges = set(TestWorkflow.get_unused_edges())

    # Expected unused edges
    expected_unused_edges = {
        Edge(from_port=NodeC.Ports.default, to_node=NodeD),
        Edge(from_port=NodeE.Ports.default, to_node=NodeF),
        Edge(from_port=NodeE.Ports.default, to_node=NodeG),
        Edge(from_port=NodeF.Ports.default, to_node=NodeH),
        Edge(from_port=NodeG.Ports.default, to_node=NodeH),
    }

    # TEST that unused edges are correctly identified
    assert unused_edges == expected_unused_edges, f"Expected {expected_unused_edges}, but got {unused_edges}"


def test_workflow__no_unused_nodes():
    class NodeA(BaseNode):
        pass

    class NodeB(BaseNode):
        pass

    # WHEN we create a workflow with no unused nodes
    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = NodeA >> NodeB

    # TEST that nodes not in the graph are empty
    nodes = set(TestWorkflow.get_unused_nodes())
    assert nodes == set()


def test_workflow__no_unused_edges():
    class NodeA(BaseNode):
        pass

    class NodeB(BaseNode):
        pass

    # WHEN we create a workflow with no unused edges
    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = NodeA >> NodeB

    # TEST that unused edges are empty
    edges = set(TestWorkflow.get_unused_edges())
    assert edges == set()


def test_workflow__node_in_both_graph_and_unused(caplog):
    class NodeA(BaseNode):
        pass

    class NodeB(BaseNode):
        pass

    class NodeC(BaseNode):
        pass

    # WHEN we create a workflow where NodeA appears in both graph and unused
    with caplog.at_level(logging.WARNING):

        class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
            graph = NodeA >> NodeB
            unused_graphs = {NodeA >> NodeC}

    # THEN it should log a warning
    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == "WARNING"
    assert (
        "Node(s) NodeA appear in both graph and unused_graphs. Removing from unused_graphs."
        in caplog.records[0].message
    )

    # AND the workflow should be created successfully
    assert TestWorkflow is not None

    # AND NodeA should be removed from unused_graphs while NodeC remains
    unused_nodes = set(TestWorkflow.get_unused_nodes())
    assert NodeA not in unused_nodes
    assert NodeC in unused_nodes


def test_workflow__unsupported_graph_item():
    with pytest.raises(TypeError) as exc_info:
        # GIVEN a workflow with an unsupported graph item
        class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
            graph = 1  # type: ignore

    # THEN it should raise an error
    assert "Unexpected graph type: <class 'int'>" in str(exc_info.value)


def test_base_workflow__deserialize_state():
    # GIVEN a state definition
    class State(BaseState):
        bar: str

    # AND an inputs definition
    class Inputs(BaseInputs):
        baz: str

    # AND a node
    class NodeA(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    # AND a node id for node A
    node_a_id = uuid4()
    NodeA.__id__ = node_a_id

    # AND an output id for NodeA.Outputs.foo
    node_a_output_id = uuid4()
    NodeA.__output_ids__["foo"] = node_a_output_id

    # AND a workflow that uses all three
    class TestWorkflow(BaseWorkflow[Inputs, State]):
        graph = NodeA

    # WHEN we deserialize the state
    last_span_id = str(uuid4())
    state = TestWorkflow.deserialize_state(
        {
            "bar": "My state bar",
            "meta": {
                "id": "b70a5a4f-8253-4a38-aeaf-0700b4783a78",
                "trace_id": "9dcfb309-81e9-4b75-9b21-edd31cf9685f",
                "span_id": "1c2e310c-3624-4f4f-b7ac-1e429de29bbf",
                "updated_ts": "2025-04-14T19:22:18.504902",
                "external_inputs": {},
                "node_outputs": {
                    str(node_a_output_id): "My node A output foo",
                },
                "node_execution_cache": {
                    "dependencies_invoked": {
                        last_span_id: [],
                    },
                    "node_executions_initiated": {
                        str(node_a_id): [last_span_id],
                    },
                    "node_executions_fulfilled": {
                        str(node_a_id): [last_span_id],
                    },
                    "node_executions_queued": {
                        str(node_a_id): [],
                    },
                },
                "parent": None,
            },
        },
        workflow_inputs=Inputs(baz="My input baz"),
    )

    # THEN the state should be correct
    assert state.bar == "My state bar"
    assert state.meta.node_outputs == {NodeA.Outputs.foo: "My node A output foo"}
    assert isinstance(state.meta.workflow_inputs, Inputs)
    assert state.meta.workflow_inputs.baz == "My input baz"

    # AND the node execution cache should deserialize correctly
    assert state.meta.node_execution_cache._node_executions_initiated == {NodeA: {UUID(last_span_id)}}
    assert state.meta.node_execution_cache._node_executions_fulfilled == {NodeA: Stack.from_list([UUID(last_span_id)])}
    assert state.meta.node_execution_cache._node_executions_queued == {NodeA: []}
    assert state.meta.node_execution_cache._dependencies_invoked == {UUID(last_span_id): set()}


def test_base_workflow__deserialize_legacy_node_execution_cache():

    # GIVEN a state definition
    class State(BaseState):
        bar: str

    # AND a node
    class NodeA(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    # AND a workflow that uses both
    class TestWorkflow(BaseWorkflow[BaseInputs, State]):
        graph = NodeA

    # WHEN we deserialize the state that had a legacy node execution cache format
    last_span_id = str(uuid4())
    node_a_full_name = "vellum.workflows.workflows.tests.test_base_workflow.test_base_workflow__deserialize_legacy_node_execution_cache.<locals>.NodeA"  # noqa: E501
    state = TestWorkflow.deserialize_state(
        {
            "meta": {
                "node_execution_cache": {
                    "dependencies_invoked": {
                        last_span_id: [],
                    },
                    "node_executions_initiated": {
                        node_a_full_name: [last_span_id],
                    },
                    "node_executions_fulfilled": {
                        node_a_full_name: [last_span_id],
                    },
                    "node_executions_queued": {
                        node_a_full_name: [],
                    },
                },
            },
        },
    )

    # THEN the node execution cache should deserialize correctly
    assert state.meta.node_execution_cache._node_executions_initiated == {NodeA: {UUID(last_span_id)}}
    assert state.meta.node_execution_cache._node_executions_fulfilled == {NodeA: Stack.from_list([UUID(last_span_id)])}
    assert state.meta.node_execution_cache._node_executions_queued == {NodeA: []}
    assert state.meta.node_execution_cache._dependencies_invoked == {UUID(last_span_id): set()}


def test_base_workflow__deserialize_legacy_node_outputs():

    # GIVEN a state definition
    class State(BaseState):
        bar: str

    # AND a node
    class NodeA(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    # AND a workflow that uses both
    class TestWorkflow(BaseWorkflow[BaseInputs, State]):
        graph = NodeA

    # WHEN we deserialize the state that had a legacy node execution cache format
    state = TestWorkflow.deserialize_state(
        {
            "meta": {
                "node_outputs": {
                    "test_base_workflow__deserialize_legacy_node_outputs.<locals>.NodeA.Outputs.foo": "My node A output foo",  # noqa: E501
                },
            },
        },
    )

    # THEN the node execution cache should deserialize correctly
    assert state.meta.node_outputs == {NodeA.Outputs.foo: "My node A output foo"}


def test_base_workflow__deserialize_state_with_optional_inputs():

    # GIVEN a state definition
    class State(BaseState):
        bar: str

    # AND an inputs definition with an optional field
    class Inputs(BaseInputs):
        baz: str = "My default baz"

    # AND a node
    class NodeA(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    # AND a workflow that uses all three
    class TestWorkflow(BaseWorkflow[Inputs, State]):
        graph = NodeA

    # WHEN we deserialize the state
    state = TestWorkflow.deserialize_state(
        {
            "bar": "My state bar",
            "meta": {
                "node_outputs": {},
                "parent": None,
            },
        },
    )

    # THEN the state should be correct
    assert state.bar == "My state bar"
    assert isinstance(state.meta.workflow_inputs, Inputs)
    assert state.meta.workflow_inputs.baz == "My default baz"


def test_base_workflow__deserialize_nested_state():
    # GIVEN a state definition
    class State(BaseState):
        foo: str

    # AND a nested state definition
    class NestedState(BaseState):
        bar: str

    # AND an inner node
    class InnerNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            baz: str

    # AND a subworkflow definition with the nested state and inner node
    class InnerWorkflow(BaseWorkflow[BaseInputs, NestedState]):
        graph = InnerNode

    # AND a subworkflow node
    class OuterNode(InlineSubworkflowNode[State, BaseInputs, NestedState]):
        subworkflow = InnerWorkflow

        class Outputs(InlineSubworkflowNode.Outputs):
            qux: str

    # AND a workflow that uses the outer state and subworkflow node
    class TestWorkflow(BaseWorkflow[BaseInputs, State]):
        graph = OuterNode

    # WHEN we deserialize the nested state
    inner_node_output_definition = "test_base_workflow__deserialize_nested_state.<locals>.InnerNode.Outputs.baz"
    outer_node_output_definition = "test_base_workflow__deserialize_nested_state.<locals>.OuterNode.Outputs.qux"
    state = InnerWorkflow.deserialize_state(
        {
            "bar": "My nested state bar",
            "meta": {
                "workflow_definition": {
                    "id": str(InnerWorkflow.__id__),
                    "name": "test_base_workflow__deserialize_nested_state.<locals>.InnerWorkflow",
                    "module": ["vellum", "workflows", "workflows", "tests", "test_base_workflow"],
                },
                "node_outputs": {inner_node_output_definition: "My inner node output baz"},
                "parent": {
                    "foo": "My outer state foo",
                    "meta": {
                        "node_outputs": {outer_node_output_definition: "My outer node output qux"},
                        "workflow_definition": {
                            "id": str(TestWorkflow.__id__),
                            "name": "test_base_workflow__deserialize_nested_state.<locals>.TestWorkflow",
                            "module": ["vellum", "workflows", "workflows", "tests", "test_base_workflow"],
                        },
                    },
                },
            },
        },
    )

    # THEN the state should be correct
    assert state.bar == "My nested state bar"
    assert state.meta.node_outputs == {InnerNode.Outputs.baz: "My inner node output baz"}
    assert state.meta.workflow_definition == InnerWorkflow

    # AND the parent state should be correct
    assert isinstance(state.meta.parent, State)
    assert state.meta.parent.foo == "My outer state foo"
    assert state.meta.parent.meta.node_outputs == {OuterNode.Outputs.qux: "My outer node output qux"}
    assert state.meta.parent.meta.workflow_definition == TestWorkflow


def test_base_workflow__deserialize_state_with_none():

    # GIVEN a state definition
    class State(BaseState):
        bar: str

    # AND an inputs definition
    class Inputs(BaseInputs):
        baz: str

    # AND a node
    class NodeA(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    # AND a workflow that uses all three
    class TestWorkflow(BaseWorkflow[Inputs, State]):
        graph = NodeA

    # WHEN we deserialize None
    state = TestWorkflow.deserialize_state(None)

    # THEN we should get back None
    assert state is None


@pytest.mark.parametrize(
    "raw_workflow_definition",
    [
        {
            "name": "test_base_workflow__deserialize_state_with_invalid_workflow_definition.<locals>.TestWorkflow",
            "module": ["invalid", "module", "path"],
        },
        {
            "name": "test_base_workflow__deserialize_state_with_invalid_workflow_definition.<locals>.InvalidWorkflow",
            "module": ["vellum", "workflows", "workflows", "tests", "test_base_workflow"],
        },
        {
            "name": "invalid_local_function_name.<locals>.TestWorkflow",
            "module": ["vellum", "workflows", "workflows", "tests", "test_base_workflow"],
        },
        {
            "name": "GlobalTestWorkflow",
            "module": ["invalid", "module", "path"],
        },
        {
            "name": "InlineSubworkflowNode",
            "module": ["vellum", "workflows", "workflows", "tests", "test_base_workflow"],
        },
        {
            "name": "InvalidWorkflow",
            "module": ["vellum", "workflows", "workflows", "tests", "test_base_workflow"],
        },
    ],
    ids=[
        "invalid_module_path_with_locals",
        "invalid_local_name",
        "invalid_local_container",
        "invalid_global_module",
        "invalid_global_name",
        "missing_global_name",
    ],
)
def test_base_workflow__deserialize_state_with_invalid_workflow_definition(raw_workflow_definition):

    # GIVEN a state definition
    class State(BaseState):
        bar = "My default bar"

    # AND a workflow
    class TestWorkflow(BaseWorkflow[BaseInputs, State]):
        pass

    # WHEN we deserialize a state with an invalid module path
    state = TestWorkflow.deserialize_state(
        {
            "meta": {
                "workflow_definition": {
                    "id": str(TestWorkflow.__id__),
                    **raw_workflow_definition,
                },
            },
        },
    )

    # THEN we should get back the default state
    assert isinstance(state, State)
    assert state.bar == "My default bar"

    # AND the workflow definition should be BaseWorkflow
    assert state.meta.workflow_definition == BaseWorkflow



import json
from typing import Sequence, Union, cast

from vellum import (
    ChatMessage,
    DocumentVellumValue,
    JsonVellumValue,
    PromptBlock,
    PromptRequestInput,
    RichTextPromptBlock,
    StringVellumValue,
    VellumDocument,
    VellumVariable,
)
from vellum.client.types.audio_vellum_value import AudioVellumValue
from vellum.client.types.function_call import FunctionCall
from vellum.client.types.function_call_vellum_value import FunctionCallVellumValue
from vellum.client.types.image_vellum_value import ImageVellumValue
from vellum.client.types.number_input import NumberInput
from vellum.client.types.vellum_audio import VellumAudio
from vellum.client.types.vellum_image import VellumImage
from vellum.prompts.blocks.exceptions import PromptCompilationError
from vellum.prompts.blocks.types import CompiledChatMessagePromptBlock, CompiledPromptBlock, CompiledValuePromptBlock
from vellum.utils.templating.constants import DEFAULT_JINJA_CUSTOM_FILTERS, DEFAULT_JINJA_GLOBALS
from vellum.utils.templating.render import render_sandboxed_jinja_template
from vellum.utils.typing import cast_not_optional

PromptInput = Union[PromptRequestInput, NumberInput]


def compile_prompt_blocks(
    blocks: list[PromptBlock],
    inputs: Sequence[PromptInput],
    input_variables: list[VellumVariable],
) -> list[CompiledPromptBlock]:
    """Compiles a list of Prompt Blocks, performing all variable substitutions and Jinja templating needed."""

    sanitized_inputs = _sanitize_inputs(inputs)
    inputs_by_name = {
        (input_.name if isinstance(input_, NumberInput) else input_.key): input_ for input_ in sanitized_inputs
    }

    compiled_blocks: list[CompiledPromptBlock] = []
    for block in blocks:
        if block.state == "DISABLED":
            continue

        if block.block_type == "CHAT_MESSAGE":
            chat_role = cast_not_optional(block.chat_role)
            inner_blocks = cast_not_optional(block.blocks)
            unterminated = block.chat_message_unterminated or False

            inner_prompt_blocks = compile_prompt_blocks(
                inner_blocks,
                sanitized_inputs,
                input_variables,
            )
            if not inner_prompt_blocks:
                continue

            compiled_blocks.append(
                CompiledChatMessagePromptBlock(
                    role=chat_role,
                    unterminated=unterminated,
                    source=block.chat_source,
                    blocks=[inner for inner in inner_prompt_blocks if inner.block_type == "VALUE"],
                    cache_config=block.cache_config,
                )
            )

        elif block.block_type == "JINJA":
            if block.template is None:
                continue

            rendered_template = render_sandboxed_jinja_template(
                template=block.template,
                input_values={name: inp.value for name, inp in inputs_by_name.items()},
                jinja_custom_filters=DEFAULT_JINJA_CUSTOM_FILTERS,
                jinja_globals=DEFAULT_JINJA_GLOBALS,
            )
            jinja_content = StringVellumValue(value=rendered_template)

            compiled_blocks.append(
                CompiledValuePromptBlock(
                    content=jinja_content,
                    cache_config=block.cache_config,
                )
            )

        elif block.block_type == "VARIABLE":
            compiled_input = inputs_by_name.get(block.input_variable)
            if compiled_input is None:
                raise PromptCompilationError(f"Input variable '{block.input_variable}' not found")

            if compiled_input.type == "CHAT_HISTORY":
                history = cast(list[ChatMessage], compiled_input.value)
                chat_message_blocks = _compile_chat_messages_as_prompt_blocks(history)
                compiled_blocks.extend(chat_message_blocks)
                continue

            if compiled_input.type == "STRING":
                compiled_blocks.append(
                    CompiledValuePromptBlock(
                        content=StringVellumValue(value=compiled_input.value),
                        cache_config=block.cache_config,
                    )
                )
            elif compiled_input.type == "JSON":
                # Skip empty JSON arrays when there are chat message blocks present
                if compiled_input.value == [] and any(block.block_type == "CHAT_MESSAGE" for block in compiled_blocks):
                    continue

                compiled_blocks.append(
                    CompiledValuePromptBlock(
                        content=JsonVellumValue(value=compiled_input.value),
                        cache_config=block.cache_config,
                    )
                )
            elif compiled_input.type == "CHAT_HISTORY":
                chat_message_blocks = _compile_chat_messages_as_prompt_blocks(compiled_input.value)
                compiled_blocks.extend(chat_message_blocks)
            else:
                raise PromptCompilationError(f"Invalid input type for variable block: {compiled_input.type}")

        elif block.block_type == "RICH_TEXT":
            value_block = _compile_rich_text_block_as_value_block(block=block, inputs=sanitized_inputs)
            compiled_blocks.append(value_block)

        elif block.block_type == "FUNCTION_DEFINITION":
            raise PromptCompilationError("Function definitions shouldn't go through compilation process")

        elif block.block_type == "FUNCTION_CALL":
            function_call_block = CompiledValuePromptBlock(
                content=FunctionCallVellumValue(
                    value=FunctionCall(
                        id=block.id,
                        name=block.name,
                        arguments=block.arguments,
                    ),
                ),
                cache_config=block.cache_config,
            )
            compiled_blocks.append(function_call_block)

        elif block.block_type == "AUDIO":
            audio_block = CompiledValuePromptBlock(
                content=AudioVellumValue(
                    value=VellumAudio(
                        src=block.src,
                        metadata=block.metadata,
                    ),
                ),
                cache_config=block.cache_config,
            )
            compiled_blocks.append(audio_block)

        # elif block.block_type == "VIDEO":
        #     video_block = CompiledValuePromptBlock(
        #         content=VideoVellumValue(
        #             value=VellumVideo(
        #                 src=block.src,
        #                 metadata=block.metadata,
        #             ),
        #         ),
        #         cache_config=block.cache_config,
        #     )
        #     compiled_blocks.append(video_block)

        elif block.block_type == "IMAGE":
            image_block = CompiledValuePromptBlock(
                content=ImageVellumValue(
                    value=VellumImage(
                        src=block.src,
                        metadata=block.metadata,
                    ),
                ),
                cache_config=block.cache_config,
            )
            compiled_blocks.append(image_block)

        elif block.block_type == "DOCUMENT":
            document_block = CompiledValuePromptBlock(
                content=DocumentVellumValue(
                    value=VellumDocument(
                        src=block.src,
                        metadata=block.metadata,
                    )
                ),
                cache_config=block.cache_config,
            )
            compiled_blocks.append(document_block)
        else:
            raise PromptCompilationError(f"Unknown block_type: {block.block_type}")

    return compiled_blocks


def _compile_chat_messages_as_prompt_blocks(chat_messages: list[ChatMessage]) -> list[CompiledChatMessagePromptBlock]:
    blocks: list[CompiledChatMessagePromptBlock] = []
    for chat_message in chat_messages:
        if chat_message.content is None:
            continue

        chat_message_blocks = (
            [
                CompiledValuePromptBlock(
                    content=item,
                )
                for item in chat_message.content.value
            ]
            if chat_message.content.type == "ARRAY"
            else [
                CompiledValuePromptBlock(
                    content=chat_message.content,
                )
            ]
        )

        blocks.append(
            CompiledChatMessagePromptBlock(
                role=chat_message.role,
                unterminated=False,
                blocks=chat_message_blocks,
                source=chat_message.source,
            )
        )

    return blocks


def _compile_rich_text_block_as_value_block(
    block: RichTextPromptBlock,
    inputs: list[PromptInput],
) -> CompiledValuePromptBlock:
    value: str = ""
    inputs_by_name = {(input_.name if isinstance(input_, NumberInput) else input_.key): input_ for input_ in inputs}
    for child_block in block.blocks:
        if child_block.block_type == "PLAIN_TEXT":
            value += child_block.text
        elif child_block.block_type == "VARIABLE":
            input = inputs_by_name.get(child_block.input_variable)
            if input is None:
                raise PromptCompilationError(f"Input variable '{child_block.input_variable}' not found")
            elif input.type == "STRING":
                value += str(input.value)
            elif input.type == "JSON":
                value += json.dumps(input.value, indent=4)
            elif input.type == "NUMBER":
                value += str(input.value)
            else:
                raise PromptCompilationError(
                    f"Input variable '{child_block.input_variable}' has an invalid type: {input.type}"
                )
        else:
            raise PromptCompilationError(f"Invalid child block_type for RICH_TEXT: {child_block.block_type}")

    return CompiledValuePromptBlock(content=StringVellumValue(value=value), cache_config=block.cache_config)


def _sanitize_inputs(inputs: Sequence[PromptInput]) -> list[PromptInput]:
    sanitized_inputs: list[PromptInput] = []
    for input_ in inputs:
        if input_.type == "CHAT_HISTORY" and input_.value is None:
            sanitized_inputs.append(input_.model_copy(update={"value": cast(list[ChatMessage], [])}))
        elif input_.type == "STRING" and input_.value is None:
            sanitized_inputs.append(input_.model_copy(update={"value": ""}))
        else:
            sanitized_inputs.append(input_)

    return sanitized_inputs



from deepdiff import DeepDiff

from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_map_node.workflow import SimpleMapExample


def test_serialize_workflow():
    # GIVEN a Workflow that uses a MapNode
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=SimpleMapExample)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 1
    assert not DeepDiff(
        [
            {
                "id": "db2eb237-38e4-417a-8bfc-5bda0f3165ca",
                "key": "fruits",
                "type": "JSON",
                "required": True,
                "default": None,
                "extensions": {"color": None},
            },
        ],
        input_variables,
        ignore_order=True,
    )

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 1
    assert not DeepDiff(
        [
            {
                "id": "145b0b68-224b-4f83-90e6-eea3457e6c3e",
                "key": "final_value",
                "type": "JSON",
            },
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    assert len(workflow_raw_data["edges"]) == 2
    assert len(workflow_raw_data["nodes"]) == 3

    # AND each node should be serialized correctly
    entrypoint_node = workflow_raw_data["nodes"][0]
    assert entrypoint_node == {
        "id": "c0aa464d-1685-4f15-a051-31b426fec92e",
        "type": "ENTRYPOINT",
        "inputs": [],
        "data": {
            "label": "Entrypoint Node",
            "source_handle_id": "844d992e-60ab-4af2-a8ff-52cd858386f7",
        },
        "base": None,
        "definition": None,
        "display_data": {
            "position": {"x": 0.0, "y": -50.0},
        },
    }

    map_node = workflow_raw_data["nodes"][1]
    assert not DeepDiff(
        {
            "id": "bf83099a-40df-4445-b90d-1f6f1067ebe3",
            "type": "MAP",
            "inputs": [
                {
                    "id": "4c0a109f-599e-4d04-a396-a51474fc2996",
                    "key": "items",
                    "value": {
                        "rules": [
                            {
                                "type": "INPUT_VARIABLE",
                                "data": {"input_variable_id": "db2eb237-38e4-417a-8bfc-5bda0f3165ca"},
                            }
                        ],
                        "combinator": "OR",
                    },
                }
            ],
            "data": {
                "label": "Map Fruits Node",
                "error_output_id": None,
                "source_handle_id": "a2171a61-0657-43ad-b6d9-cf93ce3270d0",
                "target_handle_id": "b5e8182e-20c5-482b-b4c5-4dde48c01472",
                "variant": "INLINE",
                "workflow_raw_data": {
                    "nodes": [
                        {
                            "id": "ff9bfe6e-839d-4d40-b8fc-313b3bbd0ab0",
                            "type": "ENTRYPOINT",
                            "inputs": [],
                            "data": {
                                "label": "Entrypoint Node",
                                "source_handle_id": "520d3616-8369-4e79-9da5-3febae299c2a",
                            },
                            "display_data": {"position": {"x": 0.0, "y": -50.0}},
                            "base": None,
                            "definition": None,
                        },
                        {
                            "id": "baf6d316-dc75-41e8-96c0-015aede96309",
                            "label": "Iteration",
                            "type": "GENERIC",
                            "display_data": {"position": {"x": 200.0, "y": -50.0}},
                            "base": {
                                "name": "BaseNode",
                                "module": ["vellum", "workflows", "nodes", "bases", "base"],
                            },
                            "definition": {
                                "name": "Iteration",
                                "module": ["tests", "workflows", "basic_map_node", "workflow"],
                            },
                            "trigger": {
                                "id": "551d5528-f4e1-42ea-bde0-9de4b4968253",
                                "merge_behavior": "AWAIT_ATTRIBUTES",
                            },
                            "ports": [
                                {"id": "71ada606-d791-4a59-a252-0795c5faeeaf", "type": "DEFAULT", "name": "default"}
                            ],
                            "adornments": None,
                            "attributes": [
                                {
                                    "id": "4306b0ce-f31c-4c8f-8122-659d40e60ffe",
                                    "name": "item",
                                    "value": {
                                        "type": "WORKFLOW_INPUT",
                                        "input_variable_id": "b29bb546-9bc8-4136-857d-8c7a464ba9d4",
                                    },
                                },
                                {
                                    "id": "bca02056-0fd9-44c6-8aa9-743dd1180d48",
                                    "name": "index",
                                    "value": {
                                        "type": "WORKFLOW_INPUT",
                                        "input_variable_id": "17e7ca49-668f-450d-a792-e1f97d13db67",
                                    },
                                },
                            ],
                            "outputs": [
                                {
                                    "id": "a7bcb362-a2b8-4476-b0de-a361efeec204",
                                    "name": "count",
                                    "type": "NUMBER",
                                    "value": None,
                                }
                            ],
                        },
                        {
                            "id": "6f4883b2-70b1-4e1c-ae15-7d0f5aec810b",
                            "type": "TERMINAL",
                            "data": {
                                "label": "Final Output",
                                "name": "count",
                                "target_handle_id": "9d74571f-b7f5-4c1d-8b7c-b9c648738a4d",
                                "output_id": "2a957315-fae0-4366-8a35-f0b315c5eade",
                                "output_type": "NUMBER",
                                "node_input_id": "05f2e4cc-0e36-4c6b-a535-be38281e5280",
                            },
                            "inputs": [
                                {
                                    "id": "05f2e4cc-0e36-4c6b-a535-be38281e5280",
                                    "key": "node_input",
                                    "value": {
                                        "rules": [
                                            {
                                                "type": "NODE_OUTPUT",
                                                "data": {
                                                    "node_id": "baf6d316-dc75-41e8-96c0-015aede96309",
                                                    "output_id": "a7bcb362-a2b8-4476-b0de-a361efeec204",
                                                },
                                            }
                                        ],
                                        "combinator": "OR",
                                    },
                                }
                            ],
                            "display_data": {"position": {"x": 400.0, "y": -50.0}},
                            "base": {
                                "name": "FinalOutputNode",
                                "module": [
                                    "vellum",
                                    "workflows",
                                    "nodes",
                                    "displayable",
                                    "final_output_node",
                                    "node",
                                ],
                            },
                            "definition": None,
                        },
                    ],
                    "edges": [
                        {
                            "id": "b59c050b-0f6a-4153-ab58-fa051222fa05",
                            "source_node_id": "ff9bfe6e-839d-4d40-b8fc-313b3bbd0ab0",
                            "source_handle_id": "520d3616-8369-4e79-9da5-3febae299c2a",
                            "target_node_id": "baf6d316-dc75-41e8-96c0-015aede96309",
                            "target_handle_id": "551d5528-f4e1-42ea-bde0-9de4b4968253",
                            "type": "DEFAULT",
                        },
                        {
                            "id": "14152688-6996-4d64-9231-a6e66a4827eb",
                            "source_node_id": "baf6d316-dc75-41e8-96c0-015aede96309",
                            "source_handle_id": "71ada606-d791-4a59-a252-0795c5faeeaf",
                            "target_node_id": "6f4883b2-70b1-4e1c-ae15-7d0f5aec810b",
                            "target_handle_id": "9d74571f-b7f5-4c1d-8b7c-b9c648738a4d",
                            "type": "DEFAULT",
                        },
                    ],
                    "display_data": {"viewport": {"x": 0.0, "y": 0.0, "zoom": 1.0}},
                    "definition": {
                        "name": "IterationSubworkflow",
                        "module": [
                            "tests",
                            "workflows",
                            "basic_map_node",
                            "workflow",
                        ],
                    },
                    "output_values": [
                        {
                            "output_variable_id": "2a957315-fae0-4366-8a35-f0b315c5eade",
                            "value": {
                                "type": "NODE_OUTPUT",
                                "node_id": "baf6d316-dc75-41e8-96c0-015aede96309",
                                "node_output_id": "a7bcb362-a2b8-4476-b0de-a361efeec204",
                            },
                        }
                    ],
                },
                "input_variables": [
                    {
                        "id": "b29bb546-9bc8-4136-857d-8c7a464ba9d4",
                        "key": "item",
                        "type": "JSON",
                        "required": True,
                        "default": None,
                        "extensions": {"color": None},
                    },
                    {
                        "id": "17e7ca49-668f-450d-a792-e1f97d13db67",
                        "key": "index",
                        "type": "NUMBER",
                        "required": True,
                        "default": None,
                        "extensions": {"color": None},
                    },
                    {
                        "id": "0acef1e7-caa1-4d9f-bb69-6981729af18d",
                        "key": "items",
                        "type": "JSON",
                        "required": True,
                        "default": None,
                        "extensions": {"color": None},
                    },
                ],
                "output_variables": [{"id": "2a957315-fae0-4366-8a35-f0b315c5eade", "key": "count", "type": "NUMBER"}],
                "concurrency": None,
                "items_input_id": "0acef1e7-caa1-4d9f-bb69-6981729af18d",
                "item_input_id": "b29bb546-9bc8-4136-857d-8c7a464ba9d4",
                "index_input_id": "17e7ca49-668f-450d-a792-e1f97d13db67",
            },
            "display_data": {"position": {"x": 200.0, "y": -50.0}},
            "base": {
                "name": "MapNode",
                "module": ["vellum", "workflows", "nodes", "core", "map_node", "node"],
            },
            "definition": {
                "name": "MapFruitsNode",
                "module": ["tests", "workflows", "basic_map_node", "workflow"],
            },
            "ports": [{"id": "a2171a61-0657-43ad-b6d9-cf93ce3270d0", "name": "default", "type": "DEFAULT"}],
        },
        map_node,
        ignore_order=True,
    )

    assert not DeepDiff(
        {
            "id": "bacc5d55-07d4-4a0a-a69e-831524480de5",
            "type": "TERMINAL",
            "data": {
                "label": "Final Output",
                "name": "final_value",
                "target_handle_id": "720dd872-2f3d-47b9-8245-89387f04f300",
                "output_id": "145b0b68-224b-4f83-90e6-eea3457e6c3e",
                "output_type": "JSON",
                "node_input_id": "57fe72e9-cc9f-48f2-b5ea-25ac52637a33",
            },
            "inputs": [
                {
                    "id": "57fe72e9-cc9f-48f2-b5ea-25ac52637a33",
                    "key": "node_input",
                    "value": {
                        "rules": [
                            {
                                "type": "NODE_OUTPUT",
                                "data": {
                                    "node_id": "bf83099a-40df-4445-b90d-1f6f1067ebe3",
                                    "output_id": "2a957315-fae0-4366-8a35-f0b315c5eade",
                                },
                            }
                        ],
                        "combinator": "OR",
                    },
                }
            ],
            "display_data": {"position": {"x": 400.0, "y": -50.0}},
            "base": {
                "module": [
                    "vellum",
                    "workflows",
                    "nodes",
                    "displayable",
                    "final_output_node",
                    "node",
                ],
                "name": "FinalOutputNode",
            },
            "definition": None,
        },
        workflow_raw_data["nodes"][2],
        # TODO: Fix output ID not referencing map node workflow output
        # https://app.shortcut.com/vellum/story/5667/fix-output-display-reference-on-map-nodes
        exclude_regex_paths=r"root\['inputs'\]\[0\]\['value'\]\['rules'\]\[0\]\['data'\]\['output_id'\]",
    )

    # AND each edge should be serialized correctly
    serialized_edges = workflow_raw_data["edges"]
    assert not DeepDiff(
        [
            {
                "id": "f39477f3-445a-412e-a8ab-371baa8ae990",
                "source_node_id": "c0aa464d-1685-4f15-a051-31b426fec92e",
                "source_handle_id": "844d992e-60ab-4af2-a8ff-52cd858386f7",
                "target_node_id": "bf83099a-40df-4445-b90d-1f6f1067ebe3",
                "target_handle_id": "b5e8182e-20c5-482b-b4c5-4dde48c01472",
                "type": "DEFAULT",
            },
            {
                "id": "47a34f6e-d139-4702-aa46-6212bb8a150f",
                "source_node_id": "bf83099a-40df-4445-b90d-1f6f1067ebe3",
                "source_handle_id": "a2171a61-0657-43ad-b6d9-cf93ce3270d0",
                "target_node_id": "bacc5d55-07d4-4a0a-a69e-831524480de5",
                "target_handle_id": "720dd872-2f3d-47b9-8245-89387f04f300",
                "type": "DEFAULT",
            },
        ],
        serialized_edges,
        ignore_order=True,
    )

    # AND the display data should be what we expect
    display_data = workflow_raw_data["display_data"]
    assert display_data == {
        "viewport": {
            "x": 0.0,
            "y": 0.0,
            "zoom": 1.0,
        }
    }

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "SimpleMapExample",
        "module": [
            "tests",
            "workflows",
            "basic_map_node",
            "workflow",
        ],
    }
